/***************************************************************************************/
/* proFit_interface.c                                                                  */
/* 													                                   */
/*                                                                                     */
/* Version history																	   */
/* 15.12.95: orig 5.0A9 release              	                                       */
/* 20.02.96: added SetWindowInfo, SetBoxTitle, InputBox for strings					   */
/* 10.03.96: added GetFunctionParamName, CancelEvent								   */
/* 15.03.96: added NumberToStr255, Str255ToNumber									   */
/* 17.03.96: added GetAndSetStatus													   */
/* 24.03.96: added SetTextFileFormat, changed SaveDataAsText						   */
/* 09.11.96: added SelectLeft, SelectTop, SelectRight, SelectBottom					   */
/* 21.12.96: added Gamma, GammaP, GammaLn											   */
/* 09.02.97: added BeginCommand, AddStringParam, AddDoubleParam, EndCommand			   */
/* 21.05.98: corrected GammaP														   */
/* 24.10.98: added GetProperty... calls, updated definition of InputRec				   */
/* 11.03.00: added SetTag/GetTag, MakeLinkedPICT									   */
/* 18.01.03: compatibility with project builder										   */
/* ??.11.03: added support for multi-valued functions								   */
/* 06.06.04: added Bessel and Airy funcitons                                           */
/* 29.07.04: updated for use with bundles and mach-o, updated for use with XCode	   */
/* 29.11.05: updated to support x86 and universal binaries							   */
/* 06.02.13: moved to Xcode 4.6, removed Carbon dependencies						   */
/* 01.07.13: migrated to new command interfaces, removed obsolete commands			   */
/***************************************************************************************/

/**************************************************************************
 This file contains code that takes care of the communication between
 an external module and proFit.
 
 
 Everything that you need to know to produce an External Module for proFit
 can be found in the file proFit_interface.h
 Just include proFit_interface.h at the beginning of the file containing your
 code, as it is done in FunctionTemplate.c and ProgramTemplate.c.
 It is easier and faster if you use one of these files to create a new external module.
 
 
 You do not need to understand what is going on in this file proFit_interface.c,
 unless you need to do very advanced and exotic things.
 **************************************************************************/










#import "proFit_interface.h"


#import <mach-o/dyld.h>
#import <dlfcn.h>









#define VERSIONNUMBER 2
/* the version number of this implementation. proFit must know that. */
/* VERSIONNUMBER was 1 prior to the release of pro Fit 6.0, which brought with it */
/* longer param names and multi-valued functions, so that the plug-in interface was updated */
/* to match the new possibilities and the version number bumped to 2. */

/* NOTE that plug-ins built with this version number will not be recognized by pro Fit versions  */
/* earlier than 6.0  */


/**************************************************************************/


/*
 The following struct is used as a parameter block to several routines
 provided by proFit.
 
 The ExtModulesParamBlock (defined in proFit_paramBlk.h) provides a pointer to
 a routine that acts as a bottleneck for your calls and that uses a pointer to the
 following struct as a parameter. The contents of the struct provide
 information on which routine to call and its parameters.
 
 You do not have to implement all this yourself. This file
 contains source code for calling all of the routines
 provided by proFit. The List of the prototypes of all these routines is given
 in proFit_interface.h and you can simply use them as they stand.
 */


#if TARGET_CPU_PPC
#pragma options align=mac68k
#endif

typedef struct
{
	OSErr	result;	/* noErr or error code */
	short	what;	/* the operation */

	/* The available values for "what" are:
	 * 12:	TestStop
	 * 17:	Input
	 * 40:	StopExecution
	 * 41:	Erf										(v. 5 and later)
	 * 42:	Erfc									(v. 5 and later)
	 * 48:  Gamma									(v. 5.1 and later)
	 * 49:  GammaLn									(v. 5.1 and later)
	 * 51:  GammaP									(v. 5.1 and later)
	 * 107: CalculateFunctionOutputs
	 
	 * All following routines are available starting from pro Fit version 5 and later.
     
	 * 181: GetColHandle
	 * 182: SetColHandle
	 * 194: DeactivateWindows
	 * 195: ActivateWindows
	 * 199:	GetAndSetStatus

	 * 301: AddStringParameterToCommand
	 * 302: AddDoubleParameterToCommand
	 * 303: EndCommand
	 * 304: BeginNamedCommand

	 * 305: GetResult

	 * 306: AddStringParameterToCommand for pointers
	 * 307: AddDoubleParameterToCommand for pointers

	 * 312: SelectDirectory
	 * 315: GetTag
	 * 316: SetTag
	 * 317: FindConstant

	 * 320: GetPropertyString_ObjectString		get string property from object designated by a string
	 * 321: GetPropertyString_ObjectID			get string property from object designated by an ID
	 * 322: GetPropertyValue_ObjectString		get value property from object designated by a string
	 * 323: GetPropertyValue_ObjectID			get value property from object designated by an ID
     
	 * 326: GetPropertyString_Object			get string property from AE object
	 * 327: GetPropertyValue_Object				get value property from AE object
     
	 * 330: EndCommandFunction (for double results)
	 * 331: EndCommandFunction (for string results)
	 * 332: EndCommandFunction (for AEDesc results)
	 * 333: EndcommandFunction (for AEDesc results, disposing destination before overwriting it, for internal progs)
	 
	 * 340: GetFunctionParameterProperties
	 * 341: SetFunctionparameterProperties
	 
	 * 223:	GetLastClickedCoordinates
	 */
	union	{
        struct	{	Boolean		retval;
        } v12;		/* executionstopped... */
        struct	{	InputRec*	r;
            short		nrArgs;
            Boolean 	retval;
        } v17;		/* MyInput */
        struct	{	SInt32		column;
            SInt32		row;
            double   	cellValue;
        } v25;		/* GetData/SetData */
        struct	{	SInt32		column;
            SInt32		row;
            Boolean		retval;
        } v28;		/* TestData */
        struct	{	void*		p[8];	/* pointers to routine parameters used by version 5.0 and later */
        } params;				/* The type of the parameters is specified in the implementation */
        /* of the routines at the end of proFit_interface.c		 */
        
	} u;
} RunTimeParamRec;
#if TARGET_CPU_PPC
#pragma options align=reset
#endif




ExtModulesParamBlock*	pbStorage;		/* a storage to hold pb as it was passed to main */


ExtModulesParamBlock* GetPb(void)
{
	return pbStorage;
}

#define CallRunTimeProc(a) pbStorage->MachORunTimeProc(a)

#ifdef __cplusplus
extern "C" {
#endif
    int mainEntry(short selector, ExtModulesParamBlock* pb);
#ifdef __cplusplus
}
#endif



int mainEntry(short selector, ExtModulesParamBlock* pb)
{
    pbStorage = pb;
    
	switch (selector) {
        case setup: 			pb->requiredGlobals = 0;
            pb->versionNumber = VERSIONNUMBER;
            if (sizeof(double) == 10) pb->codeType = CPU68noFPU;
            else if (sizeof(double) == 12) pb->codeType = CPU68FPU;
#if TARGET_CPU_PPC
            else pb->codeType = CPUPowerPC;
#elif TARGET_CPU_X86
            else pb->codeType = CPUx86;
#endif
            SetUp(&pb->moduleKind,pb->name,&pb->requiredGlobals,pb);
            break;
        case funcInitialize:	pb->hasDerivatives = 0;
            InitializeFunc(&pb->hasDerivatives,
                           pb->descr1,pb->descr2,&pb->numberOfParams,
                           &pb->a0,pb);
            break;
        case progInitialize: 	InitializeProg(pb); break;
        case progRun: 			Run(pb); break;
        case funcCheck: 		pb->answer=Check(pb->paramNo,&pb->a0,pb); break;
        case funcFirst: 		First(*pb->a,pb); break;
        case funcFunc: 			Func(*pb->x,*pb->a,pb->y,pb); break;
        case funcDerivatives:	Derivatives(*pb->x,*pb->a,*pb->dyda,pb); break;
        case funcLast: 			Last(pb); break;
        case cleanUp: 			CleanUp(pb); break;
        default: break;
	}
    
	return 0;
}


/* utility function to copy one string to another string.
 The parameter "limit" gives the maximal length */

void SetPascalStr(Str255 s1, const Str255 s2, short limit)
{
	if (s2[0] < limit)
		limit = s2[0];
	if (s1)
	{
        memmove(s1, s2, limit+1);
		s1[0] = limit;
	}
}



/********************************************************************************************/
/*	 implementation of the functions provided by pro Fit and listed in proFit_interface.h

 IMPORTANT:
 
 DON'T CHANGE THE TYPES OF THE FUNCTION PARAMETERS, BECAUSE MANY ROUTINES SET DIRECTLY
 POINTERS TO THIS PARAMETERS IN THE PARAMETER BLOCK USED BY pro Fit.
 
 pro Fit EXPECTS THESE POINTERS TO POINT TO PARAMETERS OF A WELL-DEFINED TYPE.
 
 IF YOU CHANGE THE TYPE OF THESE PARAMETERS, YOUR CODE WILL NOT WORK ANYMORE WITH
 PROFIT AND YOU WILL EXPERIENCE THE MOST HORRIBLE CRASHES!
 
 If you need to change some function-parameter types because of your particular
 compiler, you still need to set the parameter block pointers to point to a parameter
 of the type specified by pro Fit.
 This is the type found in this interface file and in the pro Fit user manual.
 
 ********************************************************************************************/



void GrMoveTo (double x, double y)
{
	BeginCommand("\pMoveTo");
	AddDoubleParam("\p", x);
	AddDoubleParam("\p", y);
	EndCommand();
}


void GrLineTo (double x, double y)
{
	BeginCommand("\pLineTo");
	AddDoubleParam("\p", x);
	AddDoubleParam("\p", y);
	EndCommand();
}


void GrMove (double x, double y)
{
	BeginCommand("\pMove");
	AddDoubleParam("\p", x);
	AddDoubleParam("\p", y);
	EndCommand();
}


void GrLine (double x, double y)
{
	BeginCommand("\pLine");
	AddDoubleParam("\p", x);
	AddDoubleParam("\p", y);
	EndCommand();
}


void GetGraphCoordinates (double* const xmin, double* const xmax, double* const ymin, double* const ymax)
{
	short	ax = GetCurrentAxis(xAxis);
	short	ay = GetCurrentAxis(yAxis);
	double	ax1 = 0, ax2 = 0, ay1 = 0, ay2 = 0;

	BeginCommand("\pGetAxisProperty");
	AddDoubleParam("\p", (double)ax);
	AddDoubleParam("\p", (double)'paFS');
	EndCommandFunctionDouble(&ax1);

	BeginCommand("\pGetAxisProperty");
	AddDoubleParam("\p", (double)ax);
	AddDoubleParam("\p", (double)'paLS');
	EndCommandFunctionDouble(&ax2);

	BeginCommand("\pGetAxisProperty");
	AddDoubleParam("\p", (double)ay);
	AddDoubleParam("\p", (double)'paFS');
	EndCommandFunctionDouble(&ay1);

	BeginCommand("\pGetAxisProperty");
	AddDoubleParam("\p", (double)ay);
	AddDoubleParam("\p", (double)'paLS');
	EndCommandFunctionDouble(&ay2);

	if (xmin)
		*xmin = ax1;
	if (xmax)
		*xmax = ax2;
	if (ymin)
		*ymin = ay1;
	if (ymax)
		*ymax = ay2;
}


void CreateNewGraph(double xmin, double xmax, double ymin, double ymax, short xScaling, short yScaling)
{
	BeginCommand("\pCreateNewGraph");
	AddDoubleParam("\p", xmin);
	AddDoubleParam("\p", xmax);
	AddDoubleParam("\p", ymin);
	AddDoubleParam("\p", ymax);
	AddDoubleParam("\p", xScaling);
	AddDoubleParam("\p", yScaling);
	EndCommand();
}


double GetRandomNumber(OSType distribution, double center, double deviation)
{
	double	d = 0;
	
	BeginCommand("\pGetRandomNumber");
	AddDoubleParam("\pdistribution", (double)distribution);
	AddDoubleParam("\pcenter", center);
	AddDoubleParam("\pdeviation", deviation);
	EndCommandFunctionDouble(&d);
	return d;
}


short TestStop(void)
{
	RunTimeParamRec pb;

	pb.what=12;
	CallRunTimeProc(&pb);
	return pb.u.v12.retval;
}


void StopExecution(void)
{
	RunTimeParamRec pb;

	pb.what=40;
	CallRunTimeProc(&pb);
}


short GetAndSetStatus(short newStatus, Str255 s)
{
	RunTimeParamRec pb;
	short retval;

	pb.what=199;
	pb.u.params.p[0]=&s[0];
	pb.u.params.p[1]=&newStatus;
	pb.u.params.p[2]=&retval;
	CallRunTimeProc(&pb);
	return retval;
}


void Write(const Str255 s)
{
	if (s[0] > 0)
	{
		CFStringRef cfs = CFStringCreateWithPascalString(kCFAllocatorDefault, s, kCFStringEncodingMacRoman);
		
		if (cfs)
		{
			WriteCFString(cfs);
			CFRelease(cfs);
		}
	}
}


void WriteCFString(CFStringRef s)
{
    RunTimeParamRec pb;
	pb.what=21500;
	pb.u.params.p[0]=&s;
	CallRunTimeProc(&pb);
}


void DrawImage(AEDesc* const info, double left, double top, double right, double bottom, void* image)
{	RunTimeParamRec pb;
    pb.what=253;
    pb.u.params.p[5]=info;
    pb.u.params.p[4]=&left;
    pb.u.params.p[3]=&top;
    pb.u.params.p[2]=&right;
    pb.u.params.p[1]=&bottom;
    pb.u.params.p[0]=image;
    CallRunTimeProc(&pb);
}		


void Writeln(const Str255 s)
{
	CFStringRef cfs = CFStringCreateWithPascalString(nil, s, kCFStringEncodingMacRoman);
    CFStringRef cfs1 = CFStringCreateWithFormat(nil, nil, CFSTR("%@\n"), cfs);
	
    if (cfs && cfs1)
    {
		WriteCFString(cfs1);
        CFRelease(cfs1);
        CFRelease(cfs);
    }
}


void WriteText(SInt32 length, unsigned char * theText)
{
	if (theText != nil)
	{
		CFStringRef cfs = CFStringCreateWithBytesNoCopy(nil, (UInt8*)theText, length,  kCFStringEncodingMacRoman, false, kCFAllocatorNull);
		
		if (cfs)
		{
			WriteCFString(cfs);
			CFRelease(cfs);
		}
	}
}


void WriteInt(SInt32 i)
{
    CFStringRef cfs = CFStringCreateWithFormat(nil, nil, CFSTR("%d"), (int)i);
	
    if (cfs)
    {
		WriteCFString(cfs);
        CFRelease(cfs);
    }
}


void WriteNumber(double d)
{
    CFStringRef cfs = CFStringCreateWithFormat(nil, nil, CFSTR("%f"), d);
	
    if (cfs)
    {
		WriteCFString(cfs);
        CFRelease(cfs);
    }
}


void NumberAsString(double x, Str255 s, Boolean minimize, short digits)
{
 	BeginCommand("\pNumberAsString");
	AddDoubleParam("\p", x);
	AddDoubleParam("\p", minimize);
	AddDoubleParam("\p", digits);
	EndCommandFunctionString(s);
}


double StringAsNumber(Str255 s)
{
	double	d = 0;
	
	BeginCommand("\pStringAsNumber");
	AddStringParam("\p", s);
	EndCommandFunctionDouble(&d);
	return d;
}


short Str255ToNumber(Str255 s, double* const x)
{
	double	d = 0;
	
	BeginCommand("\pStringAsNumber");
	AddStringParam("\p", s);
	AddDoublePtrParam("\p", &d);
	EndCommandFunctionDouble(x);
	return (short)rint(d);
}


short AlertBox(const Str255 message)
// shows an alert box with a message m
// returns 0 if user pressed ok, 1 if user pressed stop
{
	double	d = 0;
	
	BeginCommand("\pAlert");
	AddStringParam("\p", message);
	EndCommandFunctionDouble(&d);
	return (short)(d - 1);	// Alert returns 1 for OK, 2 for Stop
}


short InputBox(short numItems, InputRec* const r)
{
	RunTimeParamRec pb;
	pb.what=17;
	pb.u.v17.nrArgs=numItems;
	pb.u.v17.r=(InputRec*)r;
	CallRunTimeProc(&pb);
	return pb.u.v17.retval;
}


short AskBox(short* const answer,const Str255 question)
// shows an alert box with a Yes and a No button and displays "message"
// answer is 0 for No and 1 for Yes or Stop
// return value is 0 if user pressed Yes or No, 1 if user pressed stop
{
	double	d = 0;
	
	BeginCommand("\pAsk");
	AddStringParam("\p", question);
	EndCommandFunctionDouble(&d);
	if (answer)
		*answer = d > 0 ? 1 : 0;
	return (d > 1) ? 1 : 0;	// Ask returns 0 for No, 1 for Yes, 2 for Stop
}


void SetColName(SInt32 col,const Str255 name)
{
	BeginCommand("\pSetColumnProperties");
	AddDoubleParam("\pcolumn", col);
	AddStringParam("\pname", name);
	EndCommand();
}


SInt32 XColumn (void)
{
	double	d = 0;
    SInt32	refID = GetCurrentWindow(dataType);
	BeginCommand("\pGetWindowProperty");
    AddDoubleParam("\p", refID);
	AddDoubleParam("\p", 'xCol');
	EndCommandFunctionDouble(&d);
	return rint(d);
}


SInt32 YColumn (void)
{
	double	d = 0;
    SInt32	refID = GetCurrentWindow(dataType);
	BeginCommand("\pGetWindowProperty");
    AddDoubleParam("\p", refID);
	AddDoubleParam("\p", 'yCol');
	EndCommandFunctionDouble(&d);
	return rint(d);
}


SInt32 XErrColumn (void)
{
	double	d = 0;
    SInt32	refID = GetCurrentWindow(dataType);
    BeginCommand("\pGetWindowProperty");
    AddDoubleParam("\p", refID);
	AddDoubleParam("\p", 'xECl');
	EndCommandFunctionDouble(&d);
	return rint(d);
}


SInt32 YErrColumn (void)
{
	double	d = 0;
    SInt32	refID = GetCurrentWindow(dataType);
    BeginCommand("\pGetWindowProperty");
    AddDoubleParam("\p", refID);
	AddDoubleParam("\p", 'yECl');
	EndCommandFunctionDouble(&d);
	return rint(d);
}


void SetDefaultCols(SInt32 xColumn, SInt32 yColumn, SInt32 xErrColumn, SInt32 yErrColumn)
{
	BeginCommand("\pSetDefaultCols");
	AddDoubleParam("\p", xColumn);
	AddDoubleParam("\p", yColumn);
	AddDoubleParam("\p", xErrColumn);
	AddDoubleParam("\p", yErrColumn);
	EndCommand();
}


Rect GetSelection(void)
{
	Rect	r;
	
	r.left = SelectLeft();
	r.top = SelectTop();
	r.right = SelectRight();
	r.bottom = SelectBottom();
	return r;
}


SInt32 SelectLeft(void)
{
	double	d = 0;
	
	BeginCommand("\pSelectLeft");
	EndCommandFunctionDouble(&d);
	return rint(d);
}


SInt32 SelectTop(void)
{
	double	d = 0;
	
	BeginCommand("\pSelectTop");
	EndCommandFunctionDouble(&d);
	return rint(d);
}


SInt32 SelectRight(void)
{
	double	d = 0;
	
	BeginCommand("\pSelectRight");
	EndCommandFunctionDouble(&d);
	return rint(d);
}


SInt32 SelectBottom(void)
{
	double	d = 0;
	
	BeginCommand("\pSelectBottom");
	EndCommandFunctionDouble(&d);
	return rint(d);
}


SInt32 NrRows(void)
{
	double	d = 0;
	
	BeginCommand("\pNrRows");
	EndCommandFunctionDouble(&d);
	return (SInt32)d;
}


SInt32 NrCols(void)
{
	double	d = 0;
	
	BeginCommand("\pNrCols");
	EndCommandFunctionDouble(&d);
	return (SInt32)d;
}


double	GetData(SInt32 row, SInt32 col)
{
    RunTimeParamRec pb;
    pb.what=25;
    pb.u.v25.row = row;
    pb.u.v25.column = col;
    CallRunTimeProc(&pb);
    return pb.u.v25.cellValue;

/*
    For performance reasons, we do not use the following
	double	d = 0;
	
	BeginCommand("\pGetData");
	AddDoubleParam("\p", row);
	AddDoubleParam("\p", col);
	EndCommandFunctionDouble(&d);
	return d;
 */
}


void SetData(SInt32 row, SInt32 col, double value)
{
    RunTimeParamRec pb;
    pb.what=26;
    pb.u.v25.row = row;
    pb.u.v25.column = col;
    pb.u.v25.cellValue = value;
    CallRunTimeProc(&pb);
/*
 For performance reasons, we do not use the following
	BeginCommand("\pSetData");
	AddDoubleParam("\p", row);
	AddDoubleParam("\p", col);
	AddDoubleParam("\p", value);
	EndCommand();
 */
}


void ClearData(SInt32 row, SInt32 col)
{
    
	BeginCommand("\pClearData");
	AddDoubleParam("\p", row);
	AddDoubleParam("\p", col);
	EndCommand();
}


short TestData(SInt32 row, SInt32 col)
{
    RunTimeParamRec pb;
    pb.what=28;
    pb.u.v28.row = row;
    pb.u.v28.column = col;
    CallRunTimeProc(&pb);
    return pb.u.v28.retval;
    
    /*
     For performance reasons, we do not use the following
	double	d = 0;
	
	BeginCommand("\pDataOK");
	AddDoubleParam("\p", row);
	AddDoubleParam("\p", col);
	EndCommandFunctionDouble(&d);
	return (short)d;
*/
}


short NumberInvalid(double aNum)
{
	return (short)isnan(aNum);
}


void SetParamLimits(short paramNum, double low, double high)
{
	BeginCommand("\pSetParameterProperties");
	AddDoubleParam("\pparameter", paramNum);
	AddDoubleParam("\pmin", low);
	AddDoubleParam("\pmax", high);
	EndCommand();
}


void SetParamDefaultValue(short paramNum, double value)
{
	BeginCommand("\pSetParameterProperties");
	AddDoubleParam("\pparameter", paramNum);
	AddDoubleParam("\pvalue", value);
	EndCommand();
}


void SetParamName(short paramNum, const Str255 name)
{
	BeginCommand("\pSetParameterProperties");
	AddDoubleParam("\pparameter", paramNum);
	AddStringParam("\pname", name);
	EndCommand();
}


void SetParamDefaults(short paramNum, double value, short mode, const Str255 name, double low, double high)
{
	BeginCommand("\pSetParameterProperties");
	AddDoubleParam("\pparameter", paramNum);
	AddDoubleParam("\pvalue", value);
	AddDoubleParam("\pmode", mode);
	AddStringParam("\pname", name);
	AddDoubleParam("\pmin", low);
	AddDoubleParam("\pmax", high);
	EndCommand();
}


void SelectFunction(const Str255 name)
{
	BeginCommand("\pSelectFunction");
	AddStringParam("\p", name);
	EndCommand();
}


double CallFunction(const Str255 name, double xvalue)
{
	double	d = 0;
	
	BeginCommand("\pCallFunction");
	AddStringParam("\p", name);
	AddDoubleParam("\p", xvalue);
	EndCommandFunctionDouble(&d);
	return d;
}


short GetNumFunctionParams(const Str255 functionName)
{
	double	d = 0;
	
	BeginCommand("\pGetFunctionProperty");
	AddStringParam("\p", functionName);
	AddDoubleParam("\p", 'nrPr');
	EndCommandFunctionDouble(&d);
	return (short)d;
}


short GetFunctionParamMode(const Str255 functionName,short paramNum)
{
	double	d = NAN;
	
	BeginCommand("\pGetFuncParamPropertyNumber");
	AddStringParam("\p", functionName);
	AddDoubleParam("\p", paramNum);
	AddDoubleParam("\p", 'mode');
	EndCommandFunctionDouble(&d);
	return (short)d;
}


void GetFunctionParamName(const Str255 functionName,short paramIndex, Str255 paramName)
{
	BeginCommand("\pGetFuncParamPropertyString");
	AddStringParam("\p", functionName);
	AddDoubleParam("\p", paramIndex);
	AddDoubleParam("\p", 'name');
	EndCommandFunctionString(paramName);
}


Boolean CalculateFunctionOutputs(const Str255 name, double x, double* params, double* const y, Boolean* const desiredOutputs)
{
	RunTimeParamRec pb;
	Boolean			retval;
	pb.what=107;
	pb.u.params.p[0]=(void*)&name[0];
	pb.u.params.p[1]=&x;
	pb.u.params.p[2]=params;
	pb.u.params.p[3]=y;
	pb.u.params.p[4]=desiredOutputs;
	pb.u.params.p[5]=&retval;
	CallRunTimeProc(&pb);
	return retval;
}


double GetFunctionParam(const Str255 functionName, short paramNum)
{
	double	d = NAN;
	
	BeginCommand("\pGetFuncParamPropertyNumber");
	AddStringParam("\p", functionName);
	AddDoubleParam("\p", paramNum);
	AddDoubleParam("\p", 'vlue');
	EndCommandFunctionDouble(&d);
	return d;
}


void SetFunctionParam(const Str255 functionName, short paramNum, double value)
{
	BeginCommand("\pSetFuncParamProperties");
	AddStringParam("\pfunction", functionName);
	AddDoubleParam("\pparameter", paramNum);
	AddDoubleParam("\pvalue", value);
	EndCommand();
}


void CallProgram(const Str255 programName)
{
	BeginCommand("\pCallProgram");
	AddStringParam("\p", programName);
	EndCommand();
}


double Integral(const Str255 functionName, double xmin, double xmax, short numIters)
{
	BeginCommand("\pAnalyze");
	AddStringParam("\pfunction", functionName);
	AddDoubleParam("\ptype", 'Inte');
	AddDoubleParam("\pxMin", xmin);
	AddDoubleParam("\pxMax", xmax);
	AddDoubleParam("\piterations", numIters);
	EndCommand();
	
	return GetResult(resIntegralValue, 0, 0);		// integral value
}


double Maximum(const Str255 functionName, double xmin, double xmax)
{
	BeginCommand("\pAnalyze");
	AddStringParam("\pfunction", functionName);
	AddDoubleParam("\ptype", 'Maxi');
	AddDoubleParam("\pxMin", xmin);
	AddDoubleParam("\pxMax", xmax);
	EndCommand();
	
	SInt32	n = GetResultLong(resExtremaCount, 0, 0);				// number of extrema found
	
	return n > 0 ? GetResult(resExtremaXValue, 1, 0) : NAN;		// x-value of the first one
}


double Minimum(const Str255 functionName, double xmin, double xmax)
{
	BeginCommand("\pAnalyze");
	AddStringParam("\pfunction", functionName);
	AddDoubleParam("\ptype", 'Mini');
	AddDoubleParam("\pxMin", xmin);
	AddDoubleParam("\pxMax", xmax);
	EndCommand();
	
	SInt32	n = GetResultLong(resExtremaCount, 0, 0);				// number of extrema found
	
	return n > 0 ? GetResult(resExtremaXValue, 1, 0) : NAN;		// x-value of the first one
}


double Root(const Str255 functionName, double xmin, double xmax)
{
	BeginCommand("\pAnalyze");
	AddStringParam("\pfunction", functionName);
	AddDoubleParam("\ptype", 'Root');
	AddDoubleParam("\pxMin", xmin);
	AddDoubleParam("\pxMax", xmax);
	EndCommand();
	
	SInt32	n = GetResultLong(resRootsCount, 0, 0);				// number of roots found
	
	return n > 0 ? GetResult(resXValue, 1, 0) : NAN;		// x-value of the first one
}


double Derivative(const Str255 functionName, double x, double scale)
// scale is obsolete
{
	BeginCommand("\pAnalyze");
	AddStringParam("\pfunction", functionName);
	AddDoubleParam("\ptype", 'Deri');
	AddDoubleParam("\pxValue", x);
	EndCommand();
	
	return GetResult(resDerivativeValue, 0, 0);		// derivative value
}


double ChiSquared(void)
{
	return GetResult(resChiSquared, 0, 0);		// chi squared
}


void Fit(const Str255 functionName, SInt32 xCol, SInt32 yCol, SInt32 errCol, double errVal, Boolean selectionOnly)
{
	// obsolete - use CurveFit or FitCreate FitSetArguments
}


void CurveFit(const Str255 functionName, OSType algorithm,
			  SInt32 xCol, SInt32 yCol,
			  SInt32 yErrCol, double yErrVal, UInt32 yErrDistribution,
			  SInt32 xErrCol, double xErrVal, UInt32 xErrDistribution,
			  Boolean selectedRowsOnly,
			  SInt32 stopCounter,
			  Boolean doErrorAnalysis,
			  double confidence, SInt32 iterations)
/*
 runs a fit for the given function and algorithm:
 algorithm: 'LeMa' = Levenburg-Marquart, 'MoCa' = MonteCarlo, 'Robu' = robust, 'Line' = linear, 'Poly' = polynomial
 xCol, yCol: the x- and y-columns
 xErrCol, yErrCol: the error column. Set to 0 if using no error or a constant / percentual error
 xErrVal, yErrVal:	if > 0: a constant error   if < 0 (-100..-0) a percentual error
 xErrDistribution, yErrDistribution: 'Gaus' = Gaussian, 'Lore' = Lorentzian, 'Expd' = exponential, 'Tuck' = Tuckey, 'Andr' = Andrew, 'Rctg' = rectangular
 selectedRowsOnly: set to true if only selected rows should be fitted
 stopCounter: maximum number of iterations for MonteCarlo fits
 doErrorAnalysis: true if error analysis is to be performed
 confidence: confidence interval for statistical error analysis
 iterations: number of iterations for statistical error analysis

 NOTE: See additional parameters in the pro Fit help book
 */
{
	OSType		xErrType = 'Err4';	// unknown error
	OSType		yErrType = 'Err4';	// unknown error

	if (xErrCol != 0)
		xErrType = 'Err1';			// individual error
	else if (xErrVal > 0)
		xErrType = 'Err2';			// constant error
	else if (xErrVal < 0)
		xErrType = 'Err3';			// percentage error

	if (yErrCol != 0)
		yErrType = 'Err1';			// individual error
	else if (yErrVal > 0)
		yErrType = 'Err2';			// constant error
	else if (yErrVal < 0)
		yErrType = 'Err3';			// percentage error

	BeginCommand("\pCreateFit");
	AddStringParam("\pfunction", functionName);
	AddDoubleParam("\palgorithm", algorithm);
	AddDoubleParam("\pxColumn", xCol);
	AddDoubleParam("\pyColumn", yCol);
	AddDoubleParam("\pxErrorKind", xErrType);
	AddDoubleParam("\pxErrorColumn", xErrCol);
	AddDoubleParam("\pxError", xErrVal < 0 ? -xErrVal : xErrVal);
	AddDoubleParam("\pxErrorDistribution", xErrDistribution);
	AddDoubleParam("\pyErrorKind", yErrType);
	AddDoubleParam("\pyErrorColumn", yErrCol);
	AddDoubleParam("\pyError", yErrVal < 0 ? -yErrVal : yErrVal);
	AddDoubleParam("\pyErrorDistribution", yErrDistribution);
	AddDoubleParam("\pselectedRowsOnly", selectedRowsOnly);
	AddDoubleParam("\pstopCounter", stopCounter);
	AddDoubleParam("\pdoErrorAnalysis", doErrorAnalysis);
	AddDoubleParam("\pconfidence", confidence);
	AddDoubleParam("\piterations", iterations);
	EndCommand();
} // CurveFit


short NumFitParams(void)
{
	return (short)GetResult(resNrFittedParameters, 0, 0);		// number of fitted parameters
}


double FittedParams(short paramNum)
{
	return GetResult(resFittedParameter, paramNum, 0);			// a parameter's fitted value
}


double ParamSD(short paramNum)
{
	return GetResult(resStandardDeviation, paramNum, 0);		// a parameter's standard deviation
}


double CovarMatrix(short paramNum1, short paramNum2)
{
	return GetResult(resCovariance, paramNum1, paramNum2);		// a value from the covariance matrix
}


double Erf(double x)
{	double retVal;
	RunTimeParamRec pb;
	pb.what=41;
	pb.u.params.p[0]=&x;
	pb.u.params.p[1]=&retVal;
	CallRunTimeProc(&pb);
	return retVal;
}

double Erfc(double x)
{	double retVal;
	RunTimeParamRec pb;
	pb.what=42;
	pb.u.params.p[0]=&x;
	pb.u.params.p[1]=&retVal;
	CallRunTimeProc(&pb);
	return retVal;
}

double Gamma(double x)
{	double retVal;
	RunTimeParamRec pb;
	pb.what=48;
	pb.u.params.p[0]=&x;
	pb.u.params.p[1]=&retVal;
	CallRunTimeProc(&pb);
	return retVal;
}
double GammaLn(double x)
{	double retVal;
	RunTimeParamRec pb;
	pb.what=49;
	pb.u.params.p[0]=&x;
	pb.u.params.p[1]=&retVal;
	CallRunTimeProc(&pb);
	return retVal;
}
double GammaP(double a, double x)
{	double retVal;
	RunTimeParamRec pb;
	pb.what=51;
	pb.u.params.p[0]=&x;
	pb.u.params.p[1]=&a;
	pb.u.params.p[2]=&retVal;
	CallRunTimeProc(&pb);
	return retVal;
}


double Bessel_J(double n, double x)
{
	double	d = 0;
	
	BeginCommand("\pBessel_J");
	AddDoubleParam("\p", n);
	AddDoubleParam("\p", x);
	EndCommandFunctionDouble(&d);
	return d;
}


double Bessel_Y(double n, double x)
{
	double	d = 0;
	
	BeginCommand("\pBessel_Y");
	AddDoubleParam("\p", n);
	AddDoubleParam("\p", x);
	EndCommandFunctionDouble(&d);
	return d;
}


double Bessel_M_I(double n, double x)
{
	double	d = 0;
	
	BeginCommand("\pBessel_M_I");
	AddDoubleParam("\p", n);
	AddDoubleParam("\p", x);
	EndCommandFunctionDouble(&d);
	return d;
}


double Bessel_M_K(double n, double x)
{
	double	d = 0;
	
	BeginCommand("\pBessel_K");
	AddDoubleParam("\p", n);
	AddDoubleParam("\p", x);
	EndCommandFunctionDouble(&d);
	return d;
}


double Bessel_S_j(double n, double x)
{
	double	d = 0;
	
	BeginCommand("\pBessel_S_j");
	AddDoubleParam("\p", n);
	AddDoubleParam("\p", x);
	EndCommandFunctionDouble(&d);
	return d;
}


double Bessel_S_y(double n, double x)
{
	double	d = 0;
	
	BeginCommand("\pBessel_S_y");
	AddDoubleParam("\p", n);
	AddDoubleParam("\p", x);
	EndCommandFunctionDouble(&d);
	return d;
}


double Airy_Ai(double x)
{
	double	d = 0;
	
	BeginCommand("\pAiry_Ai");
	AddDoubleParam("\p", x);
	EndCommandFunctionDouble(&d);
	return d;
}


double Airy_Bi(double x)
{
	double	d = 0;
	
	BeginCommand("\pAiry_Bi");
	AddDoubleParam("\p", x);
	EndCommandFunctionDouble(&d);
	return d;
}


void BeginCommand(Str255 const commandName)
{
	RunTimeParamRec pb;
	pb.what=304;
	pb.u.params.p[0]=(void*)&commandName[0];
	CallRunTimeProc(&pb);
}


void AddStringParam(Str255 const paramName, Str255 const param)
{
	RunTimeParamRec pb;
	pb.what=301;
	pb.u.params.p[0]=(void*)&param[0];
	if (paramName==nil || paramName[0] == 0) pb.u.params.p[1]=nil;
	else pb.u.params.p[1]=(void*)&paramName[0];
	CallRunTimeProc(&pb);
}


void AddDoubleParam(Str255 const paramName, double param)
{
	RunTimeParamRec pb;
	pb.what=302;
	pb.u.params.p[0]=(void*)&param;
	if (paramName==nil || paramName[0] == 0) pb.u.params.p[1]=nil;
	else pb.u.params.p[1]=(void*)&paramName[0];
	CallRunTimeProc(&pb);
}


void AddStringPtrParam(Str255 const paramName, Str255* param)
{
	RunTimeParamRec pb;
	pb.what=301;
	pb.u.params.p[0]=param;
	if (paramName==nil || paramName[0] == 0) pb.u.params.p[1]=nil;
	else pb.u.params.p[1]=(void*)&paramName[0];
	CallRunTimeProc(&pb);
}


void AddCFStringParam(Str255 const paramName, CFStringRef param)
{
	RunTimeParamRec pb;
	pb.what=350;
	pb.u.params.p[0]=(void*)param;
	if (paramName==nil || paramName[0] == 0) pb.u.params.p[1]=nil;
	else pb.u.params.p[1]=(void*)&paramName[0];
	CallRunTimeProc(&pb);
}


void AddAEDescParam(Str255 const paramName, AEDesc* const param)
{
	RunTimeParamRec pb;
	pb.what=318;
	pb.u.params.p[0]=(void*)param;
	if (paramName==nil || paramName[0] == 0) pb.u.params.p[1]=nil;
	else pb.u.params.p[1]=(void*)&paramName[0];
	CallRunTimeProc(&pb);
}


void AddDoublePtrParam(Str255 const paramName, double* param)
{
	RunTimeParamRec pb;
	pb.what=302;
	pb.u.params.p[0]=param;
	if (paramName==nil || paramName[0] == 0) pb.u.params.p[1]=nil;
	else pb.u.params.p[1]=(void*)&paramName[0];
	CallRunTimeProc(&pb);
}


void EndCommand(void)
{
	RunTimeParamRec pb;
	pb.what=303;
	CallRunTimeProc(&pb);
}


void EndCommandFunction(AEDesc* const result)
{
	RunTimeParamRec pb;
	pb.what=332;
	pb.u.params.p[0]=result;
	CallRunTimeProc(&pb);
}


void EndCommandFunctionString(Str255 s)
{
	RunTimeParamRec pb;
	pb.what=331;
	pb.u.params.p[0]=s;
	CallRunTimeProc(&pb);
}


void EndCommandFunctionDouble(double* const x)
{
	RunTimeParamRec pb;
	pb.what=330;
	pb.u.params.p[0]=x;
	CallRunTimeProc(&pb);
}


static void GetResultDeep(ResultCode what, SInt32 i1, SInt32 i2, void* const p)
{
	RunTimeParamRec pb;
	SInt32			two = 2;
	
	pb.what = 305;
	pb.u.params.p[0]=&what;
	pb.u.params.p[1]=&two;
	pb.u.params.p[2]=&i2;
	pb.u.params.p[3]=&i1;
	pb.u.params.p[4]=p;
	CallRunTimeProc(&pb);
}


double GetResult(ResultCode what, SInt32 i1, SInt32 i2)
{
	double			retVal;
	
	GetResultDeep(what, i1, i2, &retVal);
	return retVal;
}


double GetResultLong(ResultCode what, SInt32 i1, SInt32 i2)
{
	SInt32			retVal;
	
	GetResultDeep(what, i1, i2, &retVal);
	return retVal;
}


void GetResultStr(ResultCode what, SInt32 i1, SInt32 i2, unsigned char* const s)
{
	GetResultDeep(what, i1, i2, s);
}


double MarkedX(short i)
{
	double	d = 0;
	
	BeginCommand("\pMarkedX");
	AddDoubleParam("\p", (double)i);
	EndCommandFunctionDouble(&d);
	return d;
}


double MarkedY(short i)
{
	double	d = 0;
	
	BeginCommand("\pMarkedY");
	AddDoubleParam("\p", (double)i);
	EndCommandFunctionDouble(&d);
	return d;
}


void MarkedCoord(short i, double* x, double* y)
{
	if (x)
		*x = MarkedX(i);
	if (y)
		*y = MarkedY(i);
}


void SetFitDefaults(short 	algorithm,
                    short 	yErrDistribution,
                    short 	xErrDistribution,
                    SInt32  	xErrColumn,
                    double  xErrValue,
                    double	stopTime)
{
	// obsolete - use CurveFit or FitCreate FitSetArguments
}


Boolean Maximize(const Str255 theFunction, double precision, Boolean varyX, double* x, double* y)
{
	Boolean retVal = true;	// we assume always positive results, as Optimize has no return value
	
	BeginCommand("\pOptimize");
	AddStringParam("\pfunction", theFunction);
	AddDoubleParam("\pprecision", precision);
	AddDoubleParam("\pvaryX", varyX);
	AddDoubleParam("\pgetMinimum", true);	// get minimum
	EndCommand();
	
	if (x)
	{
		BeginCommand("\pGetResult");
		AddDoubleParam("\p", '00oX');
		EndCommandFunctionDouble(x);
	}
	if (y)
	{
		BeginCommand("\pGetResult");
		AddDoubleParam("\p", '00oY');
		EndCommandFunctionDouble(y);
	}
	return retVal;
}


Boolean Minimize(const Str255 theFunction, double precision, Boolean varyX, double*	x, double* y)
{
	Boolean retVal = true;	// we assume always positive results, as Optimize has no return value
	
	BeginCommand("\pOptimize");
	AddStringParam("\pfunction", theFunction);
	AddDoubleParam("\pprecision", precision);
	AddDoubleParam("\pvaryX", varyX);
	AddDoubleParam("\pgetMinimum", false);	// get maximum
	EndCommand();
	
	if (x)
	{
		BeginCommand("\pGetResult");
		AddDoubleParam("\p", '00oX');
		EndCommandFunctionDouble(x);
	}
	if (y)
	{
		BeginCommand("\pGetResult");
		AddDoubleParam("\p", '00oY');
		EndCommandFunctionDouble(y);
	}
	return retVal;
}


void SetErrorAnalysis(double confidence, SInt32 iterations)
{
	// obsolete - use CurveFit or FitCreate FitSetArguments
}


void ConfidenceInterval(short i, double* const min, double* const max)
{
	if (min)
		*min = GetResult(resConfidenceMin, i, 0);
	if (max)
		*max = GetResult(resConfidenceMax, i, 0);
}


Boolean CalcStat(SInt32 column,
                 Boolean selRowsOnly,
                 Boolean withBasics,
                 Boolean withSkewAndCurt,
                 Boolean withMedian)
{
	BeginCommand("\pStatistics");
	AddDoubleParam("\pcolumn", (double)column);
	AddDoubleParam("\pselectedRowsOnly", (double)selRowsOnly);
	AddDoubleParam("\pwithBasic", (double)withBasics);
	AddDoubleParam("\pwithSkew", (double)withSkewAndCurt);
	AddDoubleParam("\pwithMedian", (double)withMedian);
	EndCommand();
	return true;
}


void GetBasics(SInt32* count,
               double* sum,
               double* mean,
               double* variance,
               double* stdDev,
               double* meanAbsDev)
{
	if (count)
		*count = (SInt32)GetResult(resStatCount, 0, 0);
	if (sum)
		*sum = GetResult(resStatSum, 0, 0);
	if (mean)
		*mean = GetResult(resStatMean, 0, 0);
	if (variance)
		*variance = GetResult(resVariance, 0, 0);
	if (stdDev)
		*stdDev = GetResult(resStatStdDeviation, 0, 0);
	if (meanAbsDev)
		*meanAbsDev = GetResult(resStatMeanAbsDeviation, 0, 0);
}


void GetSkewAndKurt(SInt32* count,
                    double* mean,
                    double* variance,
                    double* skewness,
                    double* kurtosis)
{
	if (count)
		*count = (SInt32)GetResult(resStatCount, 0, 0);
	if (mean)
		*mean = GetResult(resStatMean, 0, 0);
	if (variance)
		*variance = GetResult(resVariance, 0, 0);
	if (skewness)
		*skewness = GetResult(resSkewness, 0, 0);
	
	if (kurtosis)
		*kurtosis = GetResult(resKurtosis, 0, 0);
}


void GetMedian(SInt32* count,
                  double* mean,
                  double* median,
                  double* minimum,
                  double* maximum)
{
	if (count)
		*count = (SInt32)GetResult(resStatCount, 0, 0);
	if (mean)
		*mean = GetResult(resStatMean, 0, 0);
	if (median)
		*median = GetResult(resStatMedian, 0, 0);
	if (minimum)
		*minimum = GetResult(resMinimum, 0, 0);
	if (maximum)
		*maximum = GetResult(resMaximum, 0, 0);
}


void DoMenu(const Str255 s)
{
	BeginCommand("\pDoMenu");
	AddStringParam("\p", s);
	EndCommand();
}


SInt32 GetFrontWindow(void)
{
	double	d = 0;
	
	BeginCommand("\pFrontWindow");
	EndCommandFunctionDouble(&d);
	return (SInt32)rint(d);
}


void DoNewWindow(SInt32 windowType)
{
	BeginCommand("\pNewWindow");
	AddDoubleParam("\ptype", (double)windowType);
	EndCommand();
}


void SetCurrentWindow(SInt32	windowID)
{
	BeginCommand("\pSetCurrentWindow");
	AddDoubleParam("\p", (double)windowID);
	EndCommand();
}


SInt32 GetCurrentWindow(SInt32 windowType)
{
	double retVal = 0;
	
	BeginCommand("\pGetCurrentWindow");
	AddDoubleParam("\p", (double)windowType);
	EndCommandFunctionDouble(&retVal);
	return (SInt32)retVal;
}


void DoCloseWindow(SInt32 windowID, Boolean saveIt)
{
	BeginCommand("\pCloseWindow");
	AddDoubleParam("\pwindow", windowID);
	AddDoubleParam("\psaveOption", saveIt ? 'yes ' : 'no  ');
	EndCommand();
}


SInt32 FrontmostWindow(SInt32 windowType)
{
	double	d = 0;
	
	BeginCommand("\pFrontmostWindow");
	AddDoubleParam("\p", windowType);
	EndCommandFunctionDouble(&d);
	return (SInt32)rint(d);
}


void SaveWindow(SInt32 windowID)
{
	BeginCommand("\pSaveWindow");
	AddDoubleParam("\pwindow", windowID);
	EndCommand();
}


void SaveWindowAs(SInt32 windowID, const Str255 fileName)
{
	BeginCommand("\pSaveWindow");
	AddDoubleParam("\pwindow", windowID);
	AddStringParam("\pfile", fileName);
	EndCommand();
}


SInt32 GetWindowID(const Str255 windowName)
{
	double	d = 0;
	
	BeginCommand("\pGetWindowID");
	AddStringParam("\p", windowName);
	EndCommandFunctionDouble(&d);
	return (SInt32)rint(d);
}


SInt32 GetWindowType(SInt32 windowID)
{
	double	d = 0;
	
	BeginCommand("\pGetWindowProperty");
	AddDoubleParam("\p", windowID);
	AddDoubleParam("\p", 'kocl');
	EndCommandFunctionDouble(&d);
	return (SInt32)rint(d);
}


SInt32 NextWindow(SInt32 windowID)
{
	double	d = 0;
	
	BeginCommand("\pNextWindow");
	AddDoubleParam("\p", windowID);
	EndCommandFunctionDouble(&d);
	return (SInt32)rint(d);
}


void OpenFile(const Str255 fileName)
{
	BeginCommand("\pOpenFile");
	AddStringParam("\pfile", fileName);
	EndCommand();
}


void OpenData(const Str255 fileName)
{
	BeginCommand("\pOpenFile");
	AddStringParam("\pfile", fileName);
	AddDoubleParam("\ptype", (double)dataType);
	EndCommand();
}


void OpenText(const Str255 fileName)
{
	BeginCommand("\pOpenFile");
	AddStringParam("\pfile", fileName);
	AddDoubleParam("\ptype", (double)textType);
	EndCommand();
}


Boolean SelectDirectory(Str255 directoryPath)
{
	BeginCommand("\pChooseDirectory");
	EndCommandFunctionString(directoryPath);
	return directoryPath[0] > 0;
}


Boolean GetFileDirectory(SInt32 windowID, Str255 directoryPath)
{
	BeginCommand("\pGetWindowProperty");
	AddDoubleParam("\p", windowID);
	AddDoubleParam("\p", 'fDir');
	EndCommandFunctionString(directoryPath);
	return directoryPath[0] > 0;
}


void SetDefaultDirectory(const Str255 directoryPath)
{
	BeginCommand("\pSetDefaultDirectory");
	AddStringParam("\p", directoryPath);
	EndCommand();
}


void DataExportOptions(OSType mode, Boolean withTitles, Boolean copyInfo, const Str255 delimiter, const Str255 terminator)
{
	BeginCommand("\pDataExportOptions");
	AddDoubleParam("\pmode", mode);
	AddDoubleParam("\pwithTitles", withTitles);
	AddDoubleParam("\pcopyInfo", copyInfo);
	AddStringParam("\pdelimiter", delimiter);
	AddStringParam("\pterminator", terminator);
	EndCommand();
}


void DataImportOptions(OSType mode, Boolean withTitles, Boolean copyInfo, const Str255 delimiter, const Str255 terminator, SInt32 headerLines)
{
	BeginCommand("\pDataImportOptions");
	AddDoubleParam("\pmode", mode);
	AddDoubleParam("\pwithTitles", withTitles);
	AddDoubleParam("\pcopyInfo", copyInfo);
	AddStringParam("\pdelimiter", delimiter);
	AddStringParam("\pterminator", terminator);
	AddDoubleParam("\pheaderLines", headerLines);
	EndCommand();
}


void SetTextFileFormat(const Str255 colDelimiter,
                       const Str255 endOfLine,
                       Boolean withColTitles,
                       Boolean copyInfoText,
                       SInt32 nrHeaderLines,
                       SInt32 inputORoutput)
{
	if (inputORoutput == 0)
		DataExportOptions('Dcus', withColTitles, copyInfoText, colDelimiter, endOfLine);
	else
		DataImportOptions('Dcus', withColTitles, copyInfoText, colDelimiter, endOfLine, nrHeaderLines);
}


void SaveDataAsText(SInt32 windowID, const Str255 fileName)
{
	BeginCommand("\pSaveWindow");
	AddDoubleParam("\pwindow", windowID);
	AddStringParam("\pfile", fileName);
	AddDoubleParam("\ptype", 'TEXT');
	EndCommand();
}


void SaveDrawingAs(SInt32 windowID, const Str255 fileName, SInt32 format)
//format:
//0 for saving the file in a standard pro Fit drawing file
//1 for saving the file as picture (having a file type PICT)	=> obsolete, not supported anymore
//2 for saving the fiel as eps-file (having a file type EPSF)
{
	OSType		tp = 'ftGF';	// default format
	
	if (format == 2)
		tp = 'EPSF';
	
	BeginCommand("\pSaveWindow");
	AddDoubleParam("\pwindow", windowID);
	AddStringParam("\pfile", fileName);
	AddDoubleParam("\ptype", (double)tp);
	EndCommand();
}


void SetWindowTitle(SInt32 windowID, const Str255 name)
{
	BeginCommand("\pSetWindowProperties");
	AddDoubleParam("\pwindow", windowID);
	AddStringParam("\pname", name);
	EndCommand();
}


void GetWindowTitle(SInt32 windowID, Str255 name)
{
	BeginCommand("\pGetWindowProperty");
	AddDoubleParam("\p", windowID);
	AddDoubleParam("\p", 'name');
	EndCommandFunctionString(name);
}


void PlaceWindow(SInt32 windowID, Rect windowRect)
{
	BeginCommand("\pSetWindowProperties");
	AddDoubleParam("\pwindow", windowID);
	AddDoubleParam("\pboundsLeft", windowRect.left);
	AddDoubleParam("\pboundsRight", windowRect.right);
	AddDoubleParam("\pboundsTop", windowRect.top);
	AddDoubleParam("\pboundsBottom", windowRect.bottom);
	EndCommand();
}


void BringWindowToFront(SInt32 windowID)
{
	BeginCommand("\pBringWindowToFront");
	AddDoubleParam("\p", windowID);
	EndCommand();
}


void Compile(SInt32 windowID)
{
	BeginCommand("\pCompile");
	AddDoubleParam("\pwindow", windowID);
	EndCommand();
}


void SetDataSize(SInt32 numberOfRows, SInt32 numberOfColumns)
{
	SInt32	refID = GetCurrentWindow(dataType);
	
	BeginCommand("\pSetWindowProperties");
	AddDoubleParam("\pwindow", refID);
	AddDoubleParam("\pnrRows", numberOfRows);
	AddDoubleParam("\pnrCols", numberOfColumns);
	EndCommand();
}


SInt32 GetColType(SInt32 columnNumber)
{
	double retVal = 0;
	
	BeginCommand("\pGetColumnProperty");
	AddDoubleParam("\p", columnNumber);
	AddDoubleParam("\p", 'type');
	EndCommandFunctionDouble(&retVal);
	return (SInt32)retVal;
}


void SetColType(SInt32 columnNumber, SInt32 theType)
{
	BeginCommand("\pSetColumnProperties");
	AddDoubleParam("\pcolumn", columnNumber);
	AddDoubleParam("\ptype", theType);
	EndCommand();
}


void SetColWidth(SInt32 columnNumber, short width)
{
	BeginCommand("\pSetColumnProperties");
	AddDoubleParam("\pcolumn", columnNumber);
	AddDoubleParam("\pwidth", width);
	EndCommand();
}


Boolean ColEmpty(SInt32 columnNumber)
{
	double retVal = 0;
	
	BeginCommand("\pGetColumnProperty");
	AddDoubleParam("\p", columnNumber);
	AddDoubleParam("\p", 'empt');
	EndCommandFunctionDouble(&retVal);
	return retVal != 0;
}


void SetBoxTitle(const Str255 title)
{
	BeginCommand("\pSetBoxTitle");
	AddStringParam("\p", title);
	EndCommand();
}


void SetWindowInfo(SInt32 windowID, SInt32 length, Ptr info)
{
	if (info != nil)
	{
		CFStringRef cfs = CFStringCreateWithBytesNoCopy(nil, (UInt8*)info, length,  kCFStringEncodingMacRoman, false, kCFAllocatorNull);
		
		if (cfs)
		{
			BeginCommand("\pSetWindowProperties");
			AddDoubleParam("\pwindow", windowID);
			AddCFStringParam("\pinfo", cfs);
			EndCommand();
			CFRelease(cfs);
		}
	}
}


void SelectCells(SInt32 left, SInt32 top, SInt32 right, SInt32 bottom)
{
	BeginCommand("\pSelectCell");
	AddDoubleParam("\pfirstColumn", left);
	AddDoubleParam("\plastColumn", right);
	AddDoubleParam("\pfirstRow", top);
	AddDoubleParam("\plastRow", bottom);
	EndCommand();
}


void SelectRows(SInt32 top, SInt32 bottom, Boolean doSelect)
{
	BeginCommand("\pSelectRow");
	AddDoubleParam("\pfirstRow", top);
	AddDoubleParam("\plastRow", bottom);
	AddDoubleParam("\poptions", doSelect ? 'SFgt' : 'SDsl');	// forget old selection or deselect it
	EndCommand();
}


Boolean RowSelected(SInt32 rowNumber)
{
	double	d = 0;
	
	BeginCommand("\pRowSelected");
	AddDoubleParam("\p", (double)rowNumber);
	EndCommandFunctionDouble(&d);
	return d != 0;
}


Boolean CellSelected(SInt32 row, SInt32 column)
{
	double	d = 0;
	
	BeginCommand("\pCellSelected");
	AddDoubleParam("\p", (double)row);
	AddDoubleParam("\p", (double)column);
	EndCommandFunctionDouble(&d);
	return d != 0;
}


void SetCell(SInt32 row,SInt32 column, const Str255 s)
{
	BeginCommand("\pSetCell");
	AddDoubleParam("\p", (double)row);
	AddDoubleParam("\p", (double)column);
	AddStringParam("\p", s);
	EndCommand();
}


void GetCell(Str255 s, SInt32 row,SInt32 column)
{
	BeginCommand("\pGetCell");
	AddDoubleParam("\p", (double)row);
	AddDoubleParam("\p", (double)column);
	EndCommandFunctionString(s);
}


void GetFunctionName(Str255 name)
{
	BeginCommand("\pGetFunctionProperty");
	AddDoubleParam("\p", 0);	// this should select the current function
	AddDoubleParam("\p", 'name');
	EndCommandFunctionString(name);
}


OSStatus GetFunctionParameterProperties(const Str255 name, SInt32 index, AERecord* const properties)
{
	RunTimeParamRec pb;
	OSStatus		retval;
    
	pb.what=340;
	pb.u.params.p[0] = &retval;
	pb.u.params.p[1]=(void*)&name[0];
	pb.u.params.p[2]=&index;
	pb.u.params.p[3]=properties;
	CallRunTimeProc(&pb);
	
	return retval;
}


OSStatus SetFunctionParameterProperties(const Str255 name, SInt32 index, AERecord* const properties)
{
	RunTimeParamRec pb;
	OSStatus		retval;
    
	pb.what=341;
	pb.u.params.p[0] = &retval;
	pb.u.params.p[1]=(void*)&name[0];
	pb.u.params.p[2]=&index;
	pb.u.params.p[3]=properties;
	CallRunTimeProc(&pb);
	
	return retval;
}


void GetColName(Str255 name, SInt32 column)
{
	BeginCommand("\pGetColumnProperty");
	AddDoubleParam("\p", column);
	AddDoubleParam("\p", 'name');
	EndCommandFunctionString(name);
}


void GetColHandle(SInt32 col, Handle* const colH, SInt32* const length,
                  SInt32* const colType, Boolean forWriting)
{
	RunTimeParamRec pb;
	pb.what=181;
	pb.u.params.p[0]=&col;
	pb.u.params.p[1]=colH;
	pb.u.params.p[2]=length;
	pb.u.params.p[3]=colType;
	pb.u.params.p[4]=&forWriting;
	CallRunTimeProc(&pb);
}


void SetColHandle(SInt32 col, Handle colH)
{
	RunTimeParamRec pb;
	pb.what=182;
	pb.u.params.p[0]=&col;
	pb.u.params.p[1]=&colH;
	CallRunTimeProc(&pb);
}


void GetDefaultData(	DoubleArrayHandle* const xColH,
                    DoubleArrayHandle* const yColH,
                    DoubleArrayHandle* const xErrColH,
                    DoubleArrayHandle* const yErrColH,
                    LongArrayHandle*   const indecesH,
                    SInt32* arraySize,
                    Boolean selectedRowsOnly,
                    DataInfo* const info)
{
	RunTimeParamRec pb;
	pb.what=183;
	pb.u.params.p[7]=xColH;
	pb.u.params.p[6]=yColH;
	pb.u.params.p[5]=xErrColH;
	pb.u.params.p[4]=yErrColH;
	pb.u.params.p[3]=indecesH;
	pb.u.params.p[2]=arraySize;
	pb.u.params.p[1]=&selectedRowsOnly;
	pb.u.params.p[0]=info;
	CallRunTimeProc(&pb);
}


void SetWaitTitle(const Str255 s1)
{
	BeginCommand("\pSetWaitTitle");
	AddStringParam("\p", s1);
	EndCommand();
}


void SetWaitText(const Str255 s1, const Str255 s2, const Str255 s3, const Str255 s4, const Str255 s5, const Str255 s6)
{
	BeginCommand("\pSetWaitText");
	AddStringParam("\p", s1);
	AddStringParam("\p", s2);
	AddStringParam("\p", s3);
	AddStringParam("\p", s4);
	AddStringParam("\p", s5);
	AddStringParam("\p", s6);
	EndCommand();
}


SInt32 CreateTextFile(const Str255 fileName, short opt)
{
	double	d = 0;
	
	BeginCommand("\pCreateTextFile");
	AddStringParam("\p", fileName);
	AddDoubleParam("\p", opt);
	EndCommandFunctionDouble(&d);
	return (SInt32)d;
}


void CloseTextFile(SInt32 fileRefNumber)
{
	BeginCommand("\pCloseTextFile");
	AddDoubleParam("\p", fileRefNumber);
	EndCommand();
}


void WriteToTextFile(SInt32 fileRefNumber)
{
	BeginCommand("\pWriteToTextFile");
	AddDoubleParam("\p", fileRefNumber);
	EndCommand();
}


void DeactivateProFitWindows(void)
{
	RunTimeParamRec pb;
	pb.what=194;
	CallRunTimeProc(&pb);
}


void ActivateProFitWindows(void)
{
	RunTimeParamRec pb;
	pb.what=195;
	CallRunTimeProc(&pb);
}

FSSpecPtr GetModuleFile(void)
{
	return &pbStorage->moduleFile;
}


void OpenCurve(const Str255 curveName)
{
	BeginCommand("\pOpenCurve");
	AddStringParam("\pname", curveName);
	//AddDoubleParam("\plegendEntry", (double)true);
	EndCommand();
}


void CloseCurve(void)
{
	BeginCommand("\pCloseCurve");
	EndCommand();
}


void OpenDataSet(short errors, Boolean connected, const Str255 datasetName)
{
	BeginCommand("\pOpenDataSet");
	AddDoubleParam("\perrors", errors);
	AddDoubleParam("\pconnected", connected);
	AddStringParam("\pname", datasetName);
	//AddDoubleParam("\plegendEntry", (double)true);
	EndCommand();
}


void CloseDataSet(void)
{
	BeginCommand("\pCloseDataSet");
	EndCommand();
}


void SetCurveFill(short whichAxis, short axisID)
{
	// use PlotData or PlotFunction instead
}


void SetEBarStyle(double caplength, double capThick, double lineThick)
{
	BeginCommand("\pSetEBarStyle");
	AddDoubleParam("\p", caplength);
	AddDoubleParam("\p", capThick);
	AddDoubleParam("\p", lineThick);
	EndCommand();
}


void AddDataPoint(double x, double y, double xErr, double yErr, double xErr1, double yErr1)
{
	BeginCommand("\pAddDataPoint");
	AddDoubleParam("\px", x);
	AddDoubleParam("\py", y);
	AddDoubleParam("\pxError", xErr);
	AddDoubleParam("\pyError", yErr);
	AddDoubleParam("\pxError2", xErr1);
	AddDoubleParam("\pyError2", yErr1);
	EndCommand();
}


double WindowSelectLeft(void)
{
	double	d = 0;
    SInt32	refID = GetCurrentWindow(dataType);
    BeginCommand("\pGetWindowProperty");
    AddDoubleParam("\p", refID);
	AddDoubleParam("\p", 'selL');
	EndCommandFunctionDouble(&d);
	return d;
}


double WindowSelectTop(void)
{
	double	d = 0;
    SInt32	refID = GetCurrentWindow(dataType);
    BeginCommand("\pGetWindowProperty");
    AddDoubleParam("\p", refID);
	AddDoubleParam("\p", 'selT');
	EndCommandFunctionDouble(&d);
	return d;
}


double WindowSelectRight(void)
{
	double	d = 0;
    SInt32	refID = GetCurrentWindow(dataType);

    BeginCommand("\pGetWindowProperty");
    AddDoubleParam("\p", refID);
	AddDoubleParam("\p", 'selR');
	EndCommandFunctionDouble(&d);
	return d;
}


double WindowSelectBottom(void)
{
	double	d = 0;
    SInt32	refID = GetCurrentWindow(dataType);
    BeginCommand("\pGetWindowProperty");
    AddDoubleParam("\p", refID);
	AddDoubleParam("\p", 'selB');
	EndCommandFunctionDouble(&d);
	return d;
}


void GetSelectionBounds(double* left, double* top, double* right, double* bottom)
{
	if (left)
		*left = WindowSelectLeft();
	if (top)
		*top = WindowSelectTop();
	if (right)
		*right = WindowSelectRight();
	if (bottom)
		*bottom = WindowSelectBottom();
}


void SetLineStyle(double thick, short dash)
{
	BeginCommand("\pSetLineStyle");
	AddDoubleParam("\pthickness", thick);
	AddDoubleParam("\pdash", dash);
	EndCommand();
}


void SetFillPattern(short pattern)
{
	BeginCommand("\pSetFillPattern");
	AddDoubleParam("\pid", pattern);
	EndCommand();
}


void SetArrowStyle(short arrowLocation, short style, double size)
{
	BeginCommand("\pSetArrowStyle");
	AddDoubleParam("\plocation", arrowLocation);
	AddDoubleParam("\pstyle", style);
	AddDoubleParam("\psize", size);
	EndCommand();
}


void SetTextStyle(const Str255 fontName, double size, short style)
{
	BeginCommand("\pSetTextStyle");
	AddStringParam("\pfontName", fontName);
	AddDoubleParam("\pstyle", style);
	AddDoubleParam("\psize", size);
	EndCommand();
}


void SetNewGraphRect(double left, double top, double right, double bottom)
{
    BeginCommand("\pSetNewGraphRect");
    AddDoubleParam("\p", left);
    AddDoubleParam("\p", top);
    AddDoubleParam("\p", right);
    AddDoubleParam("\p", bottom);
    EndCommand();
}


void SetDataPointStyle(short style, double size, double thickness)
{
	BeginCommand("\pSetDataPointStyle");
	AddDoubleParam("\pstyle", style);
	AddDoubleParam("\psize", size);
	AddDoubleParam("\pthickness", thickness);
	EndCommand();
}


void SetBGDataPointStyle(short style, double size)
{
	BeginCommand("\pSetBGDataPointStyle");
	AddDoubleParam("\pstyle", style);
	AddDoubleParam("\psize", size);
	EndCommand();
}


void SetLineColor(SInt32 red, SInt32 green, SInt32 blue)
{
	BeginCommand("\pSetLineColor");
	AddDoubleParam("\pred", red);
	AddDoubleParam("\pgreen", green);
	AddDoubleParam("\pblue", blue);
	EndCommand();
}


void SetFillColor(SInt32 red, SInt32 green, SInt32 blue)
{
	BeginCommand("\pSetFillColor");
	AddDoubleParam("\pred", red);
	AddDoubleParam("\pgreen", green);
	AddDoubleParam("\pblue", blue);
	EndCommand();
}


void DrawRect(double left, double top, double right, double bottom)
{
	BeginCommand("\pNewShape");
	AddDoubleParam("\pshapeClass", kRectangleShape);
	AddDoubleParam("\pxPosition", (left + right) / 2);
	AddDoubleParam("\pyPosition", (top + bottom) / 2);
	AddDoubleParam("\pxSize", right - left);
	AddDoubleParam("\pySize", bottom - top);
	EndCommand();
}


void DrawArc(double left, double top, double right, double bottom, double arcStart, double arcLength)
{
	BeginCommand("\pNewShape");
	AddDoubleParam("\pshapeClass", kOvalShape);
	AddDoubleParam("\pxPosition", (left + right) / 2);
	AddDoubleParam("\pyPosition", (top + bottom) / 2);
	AddDoubleParam("\pxSize", right - left);
	AddDoubleParam("\pySize", bottom - top);
	AddDoubleParam("\parcStart", arcStart);
	AddDoubleParam("\parcLength", arcLength);
	EndCommand();
}


void DrawEllipse(double left, double top, double right, double bottom)
{
	BeginCommand("\pNewShape");
	AddDoubleParam("\pshapeClass", kOvalShape);
	AddDoubleParam("\pxPosition", (left + right) / 2);
	AddDoubleParam("\pyPosition", (top + bottom) / 2);
	AddDoubleParam("\pxSize", right - left);
	AddDoubleParam("\pySize", bottom - top);
	EndCommand();
}


void DrawLine(double start_h, double start_v, double end_h, double end_v)
{
	BeginCommand("\pNewShape");
	AddDoubleParam("\pshapeClass", kLineShape);
	AddDoubleParam("\pxPosition", start_h);
	AddDoubleParam("\pyPosition", start_v);
	AddDoubleParam("\pxSize", end_h - start_h);
	AddDoubleParam("\pySize", end_v - start_v);
	EndCommand();
}


void DrawTextLine(const Str255 theString, double theAngle, Boolean docenter)
{
	BeginCommand("\pDrawText");
	AddStringParam("\p", theString);
	AddDoubleParam("\p", theAngle);
	AddDoubleParam("\p", docenter ? 0.0 : -1.0);
	EndCommand();
}


void DrawNumber(double theNum, short decs, double theAngle, Boolean docenter)
{
	BeginCommand("\pDrawNumber");
	AddDoubleParam("\p", theNum);
	AddDoubleParam("\p", decs);
	AddDoubleParam("\p", theAngle);
	AddDoubleParam("\p", docenter ? 0.0 : -1.0);
	EndCommand();
}


void DrawDataPoint(double x, double y)
{
	BeginCommand("\pNewShape");
	AddDoubleParam("\pshapeClass", kPointShape);
	AddDoubleParam("\pxPosition", x);
	AddDoubleParam("\pyPosition", y);
	EndCommand();
}


void SetTag(const Str255 window, const Str255 program, const Str255 name, AEDesc* const desc, Boolean permanent)
{
	RunTimeParamRec pb;
	pb.what=316;
	pb.u.params.p[4]=(void*)window;
	pb.u.params.p[3]=(void*)program;
	pb.u.params.p[2]=(void*)name;
	pb.u.params.p[1]=desc;
	pb.u.params.p[0]=&permanent;
	CallRunTimeProc(&pb);
}

void GetTag(const Str255 window, const Str255 program, const Str255 name, AEDesc* const desc)
{
	RunTimeParamRec pb;
	pb.what=315;
	pb.u.params.p[3]=(void*)window;
	pb.u.params.p[2]=(void*)program;
	pb.u.params.p[1]=(void*)name;
	pb.u.params.p[0]=desc;
	CallRunTimeProc(&pb);
}

double FindConstant(const Str255 name)
{
	double	retval;
    
	RunTimeParamRec pb;
	pb.what=317;
	pb.u.params.p[1]=&retval;
	pb.u.params.p[0]=(void*)name;
	CallRunTimeProc(&pb);
	return retval;
}


void ClickedX(double* x)
{
	BeginCommand("\pClickedX");
	EndCommandFunctionDouble(x);
}


void ClickedY(double* y)
{
	BeginCommand("\pClickedY");
	EndCommandFunctionDouble(y);
}


void GetLastClickedCoordinates(double* x, double* y)
{
	ClickedX(x);
	ClickedY(y);
}


void OpenPolygon(short smoothing, short kind)
{
	BeginCommand("\pOpenPoly");
	AddDoubleParam("\p", smoothing);
	AddDoubleParam("\p", kind);
	EndCommand();
}


void ClosePolygon(void)
{
	BeginCommand("\pClosePoly");
	EndCommand();
}


void GroupBegin(void)
{
	BeginCommand("\pGroupBegin");
	EndCommand();
}


void GroupEnd(void)
{
	BeginCommand("\pGroupEnd");
	EndCommand();
}


void SetCurrentGraph(SInt32 graphID)
{
	BeginCommand("\pSetCurrentGraph");
	AddDoubleParam("\p", (double)graphID);
	EndCommand();
}


SInt32 GetCurrentGraph(void)
{
	double	d = 0;
	
	BeginCommand("\pGetCurrentGraph");
	EndCommandFunctionDouble(&d);
	return (SInt32)d;
}


SInt32 GetNextGraph(SInt32 graphID)
{
	double	d = 0;
	
	BeginCommand("\pGetCurrentGraph");
	AddDoubleParam("\p", (double)graphID);
	EndCommandFunctionDouble(&d);
	return (SInt32)d;
}


void GetGraphFrame(double* left, double* top, double* right, double* bottom)
{
	SInt32	gID = GetCurrentGraph();
	double	l = 0, t = 0, h = 0, w = 0;

	if (gID == 0)
		return;
	BeginCommand("\pGetShapeProperty");
	AddDoubleParam("\p", (double)gID);
	AddDoubleParam("\p", (double)'pXps');
	EndCommandFunctionDouble(&l);

	BeginCommand("\pGetShapeProperty");
	AddDoubleParam("\p", (double)gID);
	AddDoubleParam("\p", (double)'pYps');
	EndCommandFunctionDouble(&t);

	BeginCommand("\pGetShapeProperty");
	AddDoubleParam("\p", (double)gID);
	AddDoubleParam("\p", (double)'pXsz');
	EndCommandFunctionDouble(&w);

	BeginCommand("\pGetShapeProperty");
	AddDoubleParam("\p", (double)gID);
	AddDoubleParam("\p", (double)'pYsz');
	EndCommandFunctionDouble(&h);

	if (left)
		*left = l;
	if (top)
		*top = t;
	if (right)
		*right = l + w;
	if (bottom)
		*bottom = t + h;
}


void SetGraphFrame(double left, double top, double right, double bottom)
{
	SInt32	gID = GetCurrentGraph();

	BeginCommand("\pSetShapeProperties");
	AddDoubleParam("\pshape", (double)gID);
	AddDoubleParam("\pxPosition", left);
	AddDoubleParam("\pyPosition", top);
	AddDoubleParam("\pxSize", right - left);
	AddDoubleParam("\pySize", bottom - top);
	EndCommand();
}


short GetCurrentAxis(short whichAxis)
{
	double	d = 0;
	
	BeginCommand("\pGetCurrentAxis");
	AddDoubleParam("\p", (double)whichAxis);
	EndCommandFunctionDouble(&d);
	return (short)d;
}


void SetCurrentAxis(short whichAxis, short axisID)
{
	BeginCommand("\pSetCurrentAxis");
	AddDoubleParam("\p", (double)whichAxis);
	AddDoubleParam("\p", (double)axisID);
	EndCommand();
}


void DeleteAxis(short whichAxis, short axisID)
{
	SInt32	axID = axisID;
	
	if (whichAxis == yAxis)
		axID += 256;
	else if (whichAxis == zAxis)
		axID += 512;
	BeginCommand("\pDeleteAxis");
	AddDoubleParam("\p", (double)axID);
	EndCommand();
}


void MakeNewAxis(short whichAxis, double min, double max, short scaling, double position)
{
	BeginCommand("\pMakeNewAxis");
	AddDoubleParam("\p", (double)whichAxis);
	AddDoubleParam("\p", (double)min);
	AddDoubleParam("\p", (double)max);
	AddDoubleParam("\p", (double)scaling);
	AddDoubleParam("\p", (double)position);
	EndCommand();
}


void SetRange(short whichAxis, double min, double max, short scaling)
{
	short	axID = GetCurrentAxis(whichAxis) - 1;
	
	if (whichAxis == yAxis)
		axID += 256;
	else if (whichAxis == zAxis)
		axID += 512;
	BeginCommand("\pSetAxisProperties");
	AddDoubleParam("\paxis", (double)axID);
	AddDoubleParam("\pfirst", min);
	AddDoubleParam("\plast", max);
	if (scaling == linScaling)
		AddDoubleParam("\pscaling", (double)'sLin');
	else if (scaling == logScaling)
		AddDoubleParam("\pscaling", (double)'sLog');
	else if (scaling == recScaling)
		AddDoubleParam("\pscaling", (double)'sRec');
	else if (scaling == probScaling)
		AddDoubleParam("\pscaling", (double)'sPrb');
	EndCommand();
}


void MakeTicks(short whichAxis, double firstMaj, double distance, short nrMinTicks)
{
	BeginCommand("\pMakeTicks");
	AddDoubleParam("\p", whichAxis);
	AddDoubleParam("\p", firstMaj);
	AddDoubleParam("\p", distance);
	AddDoubleParam("\p", nrMinTicks);
	EndCommand();
}


void SetLabelsFormat(short whichAxis, short format, short decimals)
{
	BeginCommand("\pSetAxisProperties");
	AddDoubleParam("\paxis", whichAxis);
	AddDoubleParam("\pformat", format == -1 ? 'Caut' : format == 0 ? 'Cdec' : format == 1 ? 'Csci' : 'Cfix');
	AddDoubleParam("\plabelsDecimals", decimals);
	EndCommand();
}


void ClearTicks(short whichAxis)
{
	BeginCommand("\pClearTicks");
	AddDoubleParam("\p", whichAxis);
	EndCommand();
}


short AddTick(short whichAxis, double tickPos, Boolean isMajor)
{
	double	ret = 0;

	BeginCommand("\pAddTick");
	AddDoubleParam("\p", whichAxis);
	AddDoubleParam("\p", tickPos);
	AddDoubleParam("\p", isMajor ? 1 : 0);
	EndCommandFunctionDouble(&ret);
	return (short)ret;
}


void ClearLabels(short whichAxis)
{
	BeginCommand("\pClearLabels");
	AddDoubleParam("\p", whichAxis);
	EndCommand();
}


void SetOneLabel(short whichAxis, short tickNum, double labelNumber)
{
	BeginCommand("\pSetLabel");
	AddDoubleParam("\p", whichAxis);
	AddDoubleParam("\p", tickNum);
	AddDoubleParam("\p", labelNumber);
	EndCommand();
}


void SetLabelText(short whichAxis, short tickNum, const Str255 labelText)
{
	BeginCommand("\pSetLabel");
	AddDoubleParam("\p", whichAxis);
	AddDoubleParam("\p", tickNum);
	AddStringParam("\p", labelText);
	EndCommand();
}


void SetAxisPosition(short whichAxis, double position)
{
	BeginCommand("\pSetAxisProperties");
	AddDoubleParam("\paxis", whichAxis);
	AddDoubleParam("\pposition", position);
	EndCommand();
}


void SetAxisAttributes(short whichAxis, SInt32 flags)
{
	SInt32		newTicksLoc = 0;	// none
	SInt32		newLabelsAtTL = 0;	// false
	SInt32		drwTicks = (flags & drawTicks);
	
	if (whichAxis == 0)
	{
		newTicksLoc =  ((flags & plusSideTicks) != 0 ? 1 : 0) + ((flags & minusSideTicks) != 0 ? 2 : 0);
		newLabelsAtTL =  ((flags & plusSideLabels) != 0 ? 1 : 0);
	}
	else
	{
		newTicksLoc =  ((flags & plusSideTicks) != 0 ? 2 : 0) + ((flags & minusSideTicks) != 0 ? 1 : 0);
		newLabelsAtTL =  ((flags & plusSideLabels) != 0 ? 0 : 1);
	}
	
	// the following takes care of the redundancy between the information in ticksLoc and the one in "drawTicks".
	if (newTicksLoc != 0)	// if a location was specified
		drwTicks = 1;		// keep it consistent
	
	BeginCommand("\pSetAxisAttributes");
	AddDoubleParam("\paxis", (double)whichAxis);
	AddDoubleParam("\psameAsMain", (double)(flags & equalToMain) != 0);
	AddDoubleParam("\pdrawLine", (double)(flags & drawAxisLine) != 0);
	AddDoubleParam("\pdrawTicks", (double)drawTicks);
	AddDoubleParam("\pdrawLabels", (double)(flags & drawMajorTickLabels) != 0);
	AddDoubleParam("\pdrawLabelsAtMinorTicks", (double)(flags & drawMinorTickLabels) != 0);
	AddDoubleParam("\pticksLocation", (double)newTicksLoc);
	AddDoubleParam("\plabelsAtTopOrLeft", (double)newLabelsAtTL);
	EndCommand();
}


void SetGraphAttributes(SInt32 flags)
{
	// plotSequence;
	// 0: Grid, Axes/Frame, Curves (old)	| 1: Grid, Curves, Axes/Frame
	// 2: Axes/Frame, Grid, Curves          | 3: Axes/Frame, Curves, Grid
	// 4: Curves, Grid, Axes/Frame          | 5: Curves, Axes/Frame, Grid
//	SInt32	cg = GetCurrentGraph();
	Boolean	behindAxes = ((flags & plotBehindAxes) != 0);
	SInt32	plotSequence = behindAxes ? 1 : 0;	// if grid in back
	
	if ((flags & gridInFront) != 0)
		plotSequence = behindAxes ? 5 : 3;	// if grid in front
	if ((flags & gridInMiddle) != 0)
		plotSequence = behindAxes ? 4 : 2;	// if grid in middle
	
	BeginCommand("\pSetShapeAttributes");
	AddDoubleParam("\pdrawFrame", (double)(flags & drawFrame) != 0);
	AddDoubleParam("\pdrawMajorVGridLines", (double)(flags & drawMajorGridX) != 0);
	AddDoubleParam("\pdrawMinorVGridLines", (double)(flags & drawMinorGridX) != 0);
	AddDoubleParam("\pdrawMajorHGridLines", (double)(flags & drawMajorGridY) != 0);
	AddDoubleParam("\pdrawMinorHGridLines", (double)(flags & drawMinorGridY) != 0);
	AddDoubleParam("\pdrawMinorHGridLines", (double)(flags & drawMinorGridY) != 0);
	AddDoubleParam("\pplotSequence", (double)plotSequence);
	EndCommand();
}

void GetPropertyStringFromString(SInt32 objectType, Str255 const object,
                                 Str255 const prop, Str255 result)
{
	RunTimeParamRec pb;
	double	type = objectType;
    
	pb.what=320;
	pb.u.params.p[3]=result;
	pb.u.params.p[2]=&type;
	pb.u.params.p[1]=(void*)&object[0];
	pb.u.params.p[0]=(void*)&prop[0];
	CallRunTimeProc(&pb);
}

void GetPropertyValueFromString(SInt32 objectType, Str255 const object,
                                Str255 const prop, double* const result)
{
	RunTimeParamRec pb;
	double	type = objectType;
    
	pb.what=322;
	pb.u.params.p[3]=result;
	pb.u.params.p[2]=&type;
	pb.u.params.p[1]=(void*)&object[0];
	pb.u.params.p[0]=(void*)&prop[0];
	CallRunTimeProc(&pb);
}



void GetPropertyStringFromID(SInt32 objectType, SInt32 objectID,
                             Str255 const prop, Str255 result)
{
	RunTimeParamRec pb;
	double	type = objectType;
	double	id = objectID;
    
	pb.what=321;
	pb.u.params.p[3]=result;
	pb.u.params.p[2]=&type;
	pb.u.params.p[1]=&id;
	pb.u.params.p[0]=(void*)&prop[0];
	CallRunTimeProc(&pb);
}


void GetPropertyValueFromID(SInt32 objectType, SInt32 objectID,
                            Str255 const prop, double* const result)	
{
	RunTimeParamRec pb;
	double	type = objectType;
	double	id = objectID;
    
	pb.what=323;
	pb.u.params.p[3]=result;
	pb.u.params.p[2]=&type;
	pb.u.params.p[1]=&id;
	pb.u.params.p[0]=(void*)&prop[0];
	CallRunTimeProc(&pb);
}

void GetPropertyStringFromObject(SInt32 objectType, AEDesc* const object,
                                 Str255 const prop, Str255 result)
{
	RunTimeParamRec pb;
	double	type = objectType;
    
	pb.what=326;
	pb.u.params.p[3]=result;
	pb.u.params.p[2]=&type;
	pb.u.params.p[1]=(void*)object;
	pb.u.params.p[0]=(void*)&prop[0];
	CallRunTimeProc(&pb);
}


void GetPropertyValueFromObject(SInt32 objectType, AEDesc* const object,
                                Str255 const prop, double* const result)	
{
	RunTimeParamRec pb;
	double	type = objectType;
    
	pb.what=327;
	pb.u.params.p[3]=result;
	pb.u.params.p[2]=&type;
	pb.u.params.p[1]=(void*)object;
	pb.u.params.p[0]=(void*)&prop[0];
	CallRunTimeProc(&pb);
}


Boolean Output(short retValIndex)
{
	double	d = 0;

	BeginCommand("\pOutput");
	AddDoubleParam("\p", (double)retValIndex);
	EndCommandFunctionDouble(&d);
	return d != 0;
}


